#define u8  unsigned char
#define u16 unsigned short
#define i8  signed char
#define i16 signed short

#define TYPE_3CH    0
#define TYPE_6CH    1
#define TYPE_9CH    2
#define TYPE_12CH   3
#define TYPE_AYX    4

typedef struct {
    u8  sign[4];        // siganture
    struct {
        u8 minor:4;
        u8 major:4;
    } ver;              // format version
    struct {
        u8 chipnum:3;   // (number of channels - 1)/3
        u8 mch_mode:1;  // multichannel playing mode: 0 - chained; 1 - independent
    } type;       
    u16 p_baseaddr;     // module base address
} pt3md_hdr;

// if pt3md_hdr.type.mch_mode == 0 then container contains ONE module
// with variable channel number, then pt3md_mdl_hdr follows immediatly.

// if pt3md_hdr.type.mch_mode == 1 then container contains SEVERAL modules,
// so 16-bit pointers to each module are located right after pt3md_hdr.
// each module processed separately from each other.

typedef struct {
    u8  pat_num;    // number of patterns
    u8  ord_num;    // number of order positions
    u8  smp_num;    // number of samples
    u8  orn_num;    // number of ornaments
    
    u8  loop_pos;   // loop position
    u8  speed;      // initial module speed
    
    u16 int_rate;   // 15625 / (int rate in hz)
    u16 clock;      // clock rate in KHz: 0x6D6 = 1,75 MHz
    
    // 16-bit pointers
    u16 pname;      // ASCIIZ module name 
    u16 pauthor;    // ASCIIZ module author
    
    u16 ppattern;   // pattern pointer table
    u16 porder;     // order table
    u16 psample;    // sample pointer table
    u16 pornament;  // ornament pointer table
    
    u16 pnotetab;   // 12 period values for first octave (C-1..B-1)
    
} pt3md_mdl_hdr;