#pragma once
#include "player.h"

// player structs
#define sgn(a)			((a) <  0  ?  -1 : ((a) >  0  ?  1  :  0)) 
#define max(a, b)		((a) > (b) ? (a) : (b))
#define min(a, b)		((a) < (b) ? (a) : (b))
#define clamp(a, l, h)	((a) > (h) ? (h) : ((a) < (l) ? (l) : (a)))

#define u8  unsigned char
#define i8  signed char
#define u16 unsigned short
#define i16 signed short

#include "pmd_events.h"

#define PT3MD_SIGN "PMD\x1A"

#pragma pack(push, 1)

typedef struct {
    char  sign[4];        //   - "PMD\x1A"
    struct {
        u8 minor:4;
        u8 major:4;
    } ver;              //   : 0x02 - v0.2
    struct {
        u8 chipnum:3;   //     
        u8 mch_mode:1;  //  :
                        // 0 -   ,   
                        //       (, sync)
                        // 1 -      
                        //     (, async)
    } type;
	u16 reserved;		// 
} pt3md_hdr;

typedef struct {
    u8  pat_num;        //   (0 - )
    u8  ord_num;        //    (0 - )
    u8  ins_num;        //   (0 - )
    u8  orn_num;        //   (0 - )
    
    u8  loop_pos;       //   
    u8  speed;          //   
    
    struct {            //  
        u8 tremor:1;    // 0 -  6xxx ()  . 1\2\3
                        // 1 -  . 1\2\3   . 6
        u8 res:5;       // 
        u8 chiptype:1;  //   (0 - AY-3-8910, 1 - YM2149F)
        u8 ayx:1;       // 1 -    AYX-32
    } flags;
    u8  reserved;       // 
    
    u16 int_rate;       // 15625 / (   )
    u16 clock;          //     : 0x6D6 = 1,75 
    
    // 
    u16 p_name;         // ASCIIZ-   (*p_name == 0 - )
    u16 p_author;       // ASCIIZ-   (*p_author == 0 - )
    
    u16 p_pattab;       //  
    u16 p_order;        //  
    u16 p_instab;       //    
    u16 p_orntab;       //    
    
    u16 p_notetab;      //   12  (little-endian)  AY
                        //    (C-1...B-1)

    u16 p_misc;         //   (*p_misc == 0 - )
} pt3md_mdl_hdr;

//  
typedef struct {
	u8 len;             //   (0 = 256!)
    u8 loop_start;      //  release- (   len-1)
	u8 linesize:7;      //      
    u8 sustain:1;       // 1 -  sustain-, 0 - 
} pt3md_ins_hdr;

typedef
    //   
	// linesize = 0 ->  !
    // linesize = 4 ->     \
    // linesize = 5 -> +  
    // linesize = 6 -> +   (  AYX-!)
	struct {
		//u16 t_ofs:12;   //  12-  
        u8  t_ofsl;

		u8  t_ofsh:4;   //       (  )
        u8  t_mask:1;   //   (0 - )
        u8  t_acc:1;    // 1 -   , 0 -  
        u8  n_acc:1;    // 1 -   , 0 -  
        u8  n_mask:1;   //   (0 - )
        
        u8  n_ofs:5;    //  5-  
        u8  v_slide:2;  // \ ,
                        // 0,1 -  , 2 - -1, 3 - +1
        u8  res1:1;     // 
        
        u8  volume:4;   //  
        u8  e_mask:1;   //   (1(!) - )
        u8  e_acc:1;    //  t_acc   
        u8  d_acc:1;    //  t_acc   
        u8  res2:1;      // 
        
        i8  e_ofs;      //  8-    ( )
        u8  d_ofs:4;    //  4-   ( )
} pt3md_ins_data;

typedef struct {
    u8 loop1_start;     //  sustain- ( )
    u8 loop1_end;       //   sustain- ( )
} pt3md_ins_ftr;

//  
typedef struct {
    u8  len;            //   (0 = 256!)
    u8  loop_start;     //   (   len-1)
} pt3md_orn;

//   
typedef struct {
	u16 p_chan[4];		//     (A->B->C->\)
} pt3md_pattern_pchan;

typedef struct {
    u8  patlen;         //   (0 == 256(!))
    
	struct {            //   (. )
        u8    len;      //    (0 == 256(!))
        u16   ofs;      //    
    } autospd;
} pt3md_pattern;

#pragma pack(pop)

// ayregs struct
struct ayregs_t {
	unsigned int val;
	bool dirty;
};

// mixer data
union mixer_t {
	struct{
		unsigned int tone:1;
		unsigned int noise:1;
		unsigned int env:1;
	};
	unsigned int val;
};

// channel structs
struct chan_t {
	unsigned int note, basenote, oldnote;
	
	struct {
		signed   int base;		// period base (note + ornament)

		signed   int slideacc;	// cmd 1\2\3 accumulator
		signed   int slide;		// after cmd 1\2\3

		signed   int ins;		// after instrument
		signed   int final;		// final period	(final = base + slide + ins)

	} tone;

	unsigned int	volume_ch;	// channel volume
	unsigned int    volume;		// final volume

	unsigned int	env;		// envelope mask
	unsigned int	globenv;	// global envelope mask (by cmd 7)

	bool            cut;		// true - note is cutted
	bool			off;		// true - note is off (goto release)

	mixer_t			mixer;		// mixer masks
	mixer_t			globmixer;	// global mixer masks (by cmd 7)

	struct {
		unsigned int num;
		unsigned int subparm;
		unsigned int parm;

		struct {
			bool active;
			bool dir;			// true - on

			unsigned int count;

			unsigned int ontime;
			unsigned int offtime;
		} tremor;

		struct {
			bool active;
			bool porta;			// true if cmd3, false if cmd1/2
			
			unsigned int portanote;

			unsigned int count;
			unsigned int delay;

			signed   int acc;
			signed   int add;
			signed   int limit; // for cmd 3
			bool         flush; // accumulator flush flag
		} slide;

		struct {
			bool active;
			
			unsigned int count;
			unsigned int delay;

			signed   int acc;
			signed   int add;
		} volslide;

	} cmd;	// for repeating commands

	struct {
		size_t num;
		pt3md_ins_hdr *hdr;
		pt3md_ins_ftr *ftr;

		unsigned int pos;
		bool sustain;

		// DYNAMIC CAST FTW \o/
		u8 *dat;
		u8 *s_start;
		u8 *s_end;
		u8 *rep;

		// data

		// tone\noise\env period accumulators
		signed   int t_acc;
		signed   int n_acc;
		signed   int e_acc;

		signed   int volbase;	// volume base

	} ins;

	struct {
		size_t num;
		pt3md_orn *hdr;
		unsigned int pos;
		i8 *rep;
		i8 *dat;
	} orn;
};

struct env_chan_t {
	struct {
		unsigned int	type;		// env type
		bool			restart;	// true - set new type
		signed   int    base;		// base period
		signed   int    acc;		// instrument accumulator
		signed   int    final;		// final
	} env;

	struct {
		unsigned int	base;		// by env chan
		unsigned int    acc;		// after instruments
		unsigned int    final;
	} noise;

};

// player internal struct
struct player_t {
	struct {
		pt3md_pattern *dat;
		unsigned int num;
		unsigned int pos;
		unsigned int ordpos;
	} pattern;

	struct {
		u8 *pchan;
		
		bool isskip;
		unsigned int skip;
		unsigned int skippos;
	} chan[4];

	struct {
		unsigned int pos;
		u8 *dat;
		u8 *rep;
	} autospd;

	u16 *instab, *orntab;
	u8  *ordtab, *pattab;

	// speeeeed
	unsigned int speed;
	unsigned int spdpos;

};

// shared resources
#define NOTES	12
#define	OCTAVES	(8+4)		// for envelope as note
extern u8 *pmdstream;
extern unsigned int notetab[NOTES * OCTAVES];
extern unsigned int voltab[16*16];
extern ayregs_t ayregs[256];
extern chan_t chan[4];
extern env_chan_t env_chan;
extern player_t player;

extern pt3md_hdr *pmdhdr;
extern pt3md_mdl_hdr *pmdmdl;

// shared procedures
void write_ay(int r, int d);
void push_ay();

void ins_set(int ch, int num, int pos);
void ins_process(int ch);

void orn_set(int ch, int num, int pos);
void orn_process(int ch);

void cmd_set(int ch, int num, int subparm, int parm);
void cmd_process(int ch);

bool pat_set(int num);
void pat_process();